""" SRT Segmentation
The simplest file format is .srt, and it can be handled pretty well with srt lib.
srtseg provides the additional feature on top of srt:
* Get duration of each video (cut in the middle of the space between two subtitle)
* Segment selection
* M3U8 export
* WebVTT export
Author: Jianshuo Wang
Last Modified: 2024-07-13
"""

import io
import json
import math


from copy import copy, deepcopy
from datetime import timedelta

import srt


class Seg:
    """Representation of a segment in video file.
    IMPORTANT:
    -- start/end: Timedelta of the start/end time based on the original video
    -- path: Relative path to the srt file. This is a base to calculate path
    -- selected: Used the mark whehter this seg is included in export
    -- subtitle: srt Subtitle object (which has its start, end, content)
    """

    start = timedelta(0)
    end = timedelta(0)
    subtitle = None
    duration = timedelta(0)
    path = ""
    selected = True

    def __repr__(self):
        """Representation of Seg

        Returns:
            str: string representation
        """
        return (
            f"Seg: {self.subtitle.index} {self.subtitle.start} "
            f"{self.subtitle.end} {self.subtitle.content}"
            f"@{self.path}"
        )


class SRTSeg:
    """Segment the video according to SRT time"""

    def __init__(self, path=None):
        """
        Initial based on srt_file using srt lib
            segements: collection of segments
            path: the SRT File
            subtitles: The subtitle collection
        Args:
            srt_file: the SRT file location
        """
        self.segments = []
        self.path = path

        if path is None:
            return

        def default_filename_func(seg):
            return f"{seg.path.replace('.srt')}/{seg.subtitle.index-1}.ts"

        self.media_url_func = default_filename_func

        subtitles = srt.parse(open(path).read())
        self.from_subtitles(subtitles, path)

    def from_subtitles(self, subtitles, path=None):
        """Build object from subtitles"""
        for sub in subtitles:
            seg = Seg()
            seg.subtitle = sub
            seg.selected = True
            seg.path = path
            self.segments.append(seg)

        self._calculate_times()
        return self

    def subs(self):
        """The subtitles of the video"""
        return [sg.subtitle for sg in self.segs()]

    def __iter__(self):
        return iter(self.segs())

    def segs(self):
        """The segments of the video. Contains subs and more information"""
        start_pointer = timedelta(0)
        results = []
        for seg in self.segments:
            if seg.selected:
                # We don't want to change the original stuff
                sg = deepcopy(seg)
                delta = sg.start - start_pointer
                sg.start -= delta
                sg.end -= delta
                sg.subtitle.start -= delta
                sg.subtitle.end -= delta
                start_pointer += sg.duration
                results.append(sg)

        return results

    def _calculate_times(self):
        segs = self.segments
        for index, _ in enumerate(segs):
            # 第一个和最后一个特殊处理
            if index == 0:
                segs[index].start = timedelta(0)
                segs[index].duration = segs[index].subtitle.end
                continue

            segs[index].start = (
                segs[index].subtitle.start + segs[index - 1].subtitle.end
            ) / 2
            segs[index - 1].end = segs[index].start
            segs[index - 1].duration = segs[index - 1].end - segs[index - 1].start

            if index == len(segs) - 1:
                segs[index].end = segs[index].subtitle.end
                segs[index].duration = segs[index].end - segs[index].start

        return self.segments

    def filter(self, lines=None, exact=True, padding=None, max_duration=None):
        """
        Filter the selected lines in the srt file
        Args:
            lines: The array of lines to be selected ordered by importance
                lines can be content, or the 0 based index of the line
                The order of the lines matters - from most important to least
            exact: Whether the match should be exact (from begining to
                    end) or substring match is OK
            ignore_case: Whether to ignore cases
            max_duration: The maxium duration of video can be returned
            padding: Select surrounding lines. Examples of pading:
                [0]: only the line
                [-1, 0, 1]: the lines before and after the selected line
                [0, 1]: the line and the following line
        Return:
            The same SRTSeg object
        """
        max_duration = max_duration or 60 * 60 * 1
        result = deepcopy(self)

        padding = padding or [0]
        duration = 0
        positives = []

        result.clear_selection()

        if lines is None:  # None lines means select all
            lines = range(len(self.segments))

        for line in lines:
            found = []

            # 支持index方式传入lines
            if isinstance(line, int):
                found.append(line)
            else:
                found = result.match_line(line, exact)

            # 检测连续性。如果数字是连续的，则不建议打断

            for item in found:
                if (
                    duration + result.segments[item].duration.total_seconds()
                    < max_duration
                ):
                    duration += result.segments[item].duration.total_seconds()
                    positives.append(item)
                    result.segments[item].selected = True
        # result.padding(padding, positives, max_duration)
        return result

    def clear_selection(self):
        """Set all segments to unselected before actions like filter"""
        for seg in self.segments:
            seg.selected = False

    def padding(self, padding, positives, max_duration):
        """Padding to include more segs than the key one"""
        duration = 0  # bugbug

        for positive in positives:
            for offset in padding:
                if offset == 0:
                    continue
                if 0 <= positive + offset <= len(self.segments) - 1:
                    duration += self.segments[
                        positive + offset
                    ].duration.total_seconds()
                    if duration > max_duration:
                        return self
                    self.segments[positive + offset].selected = True
        return self

    def selected_index(self):
        """Return selected index"""
        # return [(sg.originalIndex, sg.subtitle.content)
        #         for sg in self.segments if sg.selected]
        return [idx for idx, sg in enumerate(self.segments) if sg.selected]

    def m3u8(self):
        """
        Return m3u8 of the current selected content

        Args:
            usepath: Use the path in the TS file location
            start: When usepath, where is the starting point of rel path
        Returns:
            str: The content of m3u8
        """
        m3u8_infs = ""
        max_duration = 0

        for seg in self.segs():
            duration = round(seg.duration.total_seconds(), 1)
            max_duration = max(duration, max_duration)

            filename = self.media_url_func(seg)

            m3u8_infs += f"""#EXT-X-DISCONTINUITY
#EXTINF:{duration},
{filename}
"""
        m3u8_header = f"""#EXTM3U
#EXT-X-VERSION:3
#EXT-X-MEDIA-SEQUENCE:0
#EXT-X-TARGETDURATION:{math.ceil(max_duration)}
"""
        m3u8_footer = """#EXT-X-ENDLIST\n"""
        return m3u8_header + m3u8_infs + m3u8_footer

    def srtstr(self):
        """Return SRT format string"""
        return srt.compose(list(self.subs()))

    def json(self, base=None):
        """Return json format of subtitles"""
        result = base or {}
        result["subtitles"] = []
        for seg in self.segs():
            result["subtitles"].append(
                {
                    "index": seg.subtitle.index - 1,
                    "start": seg.subtitle.start.total_seconds(),
                    "end": seg.subtitle.end.total_seconds(),
                    "content": seg.subtitle.content,
                    "filename": self.media_url_func(seg),
                }
            )

        return json.dumps(result, indent=2, ensure_ascii=False)

    def __add__(self, other):
        """Append a SRT to the end of the other SRT."""
        assert issubclass(other.__class__, SRTSeg), "Not SRTSeg"
        result = copy(self)
        result.segments = copy(self.segments)

        # For everything in the append, the timestamp in everything
        # should be reset by the delta
        self_seg = list(self.segs())
        prev_end = timedelta(0) if len(self_seg) == 0 else self_seg[-1].end
        new_seg = list(other.segs())
        next_start = timedelta(0) if len(new_seg) == 0 else new_seg[0].start

        # 添加的需要都进行时间移动
        delta = prev_end - next_start
        for seg in other.segs():
            seg.start += delta
            seg.end += delta
            seg.subtitle.start += delta
            seg.subtitle.end += delta
            result.segments.append(seg)

        return result

    def __repr__(self):
        """Representation of SRT Object"""
        return (
            f"{self.__class__.__name__} filename={self.srtfile}"
            f" <{len(self.segs())}/{len(self.segments)}>"
        )



from srtseg import SRTSeg
import os


def ffmpegcmd(src="", dst="", inputcmd="", outputcmd="", use_temp=True):
    """Run ffmpeg command"""
    if os.path.exists(dst):
        return dst

    if src != dst:
        tpf = temp_file(dst)
        dst_str = f'"{tpf}" && mv "{tpf}" "{dst}"' if use_temp else f'"{dst}"'
        cmd = f'ffmpeg -hide_banner -threads 16 {inputcmd} -i "{src}" '
        cmd += f"{outputcmd} {dst_str}"
        lprint(f"[+] RUNNING {cmd}", src)
        run_as_subprocess(cmd, src)
        if not os.path.exists(dst):
            raise FileNotFoundError(f"[-]ffmpeg生成失败src: {src} cmd: {cmd}")
    return dst


def cut(src, srt_file=None):
    """把videofile按照它的字幕文件分割成很多小块
    Args:
        src 视频文件名。
        srt_file: If not provided, it implies the srt file is at
            the same location with the video file
    Returns:
        dst: The first file path of the cut segment
    """

    # 准备好文件名
    title, ext = os.path.splitext(src)
    if not srt_file:
        srt_file = title + ".srt"

    segs = SRTSeg(srt_file).segs()

    # 所有ts文件放在同名文件夹里面
    os.makedirs(title, 0o777, exist_ok=True)

    times = []
    for seg in segs:
        start = seg.start.total_seconds()
        if start == 0:
            continue
        times.append(str(start))

    last = len(times) - 1
    beacon_file = f"{title}/{last}.ts"
    print(f"[+] 寻找缓存标志文件 {beacon_file}")
    if os.path.exists(beacon_file):
        print("[*] 找到切分好的缓存文件，跳过此步骤 100%")
        return f"{title}/0.ts"

    timesstr = ",".join(times)
    options = f' -vf "scale=480:-2,fps=24" -crf 19 -c:v h264 -c:a aac  \
     -pix_fmt yuv420p  -ac 2 \
     -map 0:v:0 -map 0:a:0 \
     -segment_format mpegts \
     -segment_list "{title}/index.m3u8" \
     -segment_list_type m3u8\
     -sc_threshold 0'
    outputcmd = (
        f"-f segment -force_key_frames {timesstr} "
        f" -segment_times {timesstr} {options} "
        " -segment_time_delta 0.05 -max_muxing_queue_size 1024 "
    )
    dst = f"{title}/%d.ts"

    # There is only one sentence. Simply convert
    if len(times) < 1:
        outputcmd = options
        dst = f"{title}/0.ts"

    common.ffmpegcmd(
        src=src, dst=dst, inputcmd="-y", outputcmd=outputcmd, use_temp=False
    )

    return f"{title}/0.ts"


if __name__ == "__main__":
    cut()



from setuptools import setup, find_packages

setup(
    name="srtseg",
    version="0.2.0",
    author="Jianshuo Wang",
    author_email="jianshuo@hotmail.com",
    description="""The simplest file format is .srt, and it can be handled pretty well with srt lib.
srtseg provides the additional feature on top of srt:
* Get duration of each video (cut in the middle of the space between two subtitle)
* Segment selection
* M3U8 export
* WebVTT export""",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jianshuo/srtseg",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



import webvtt
from datetime import timedelta
from srtseg import Seg


def srt_to_vtt(srtstr):
    """Convert SRT to VTT file"""
    vtt = webvtt.parsers.SRTParser().read_from_buffer(io.StringIO(srtstr))
    for index, caption in enumerate(vtt.captions):
        caption.identifier = "s" + str(index)
    return webvtt.writers.WebVTTWriter().webvtt_content(vtt.captions)


def seg_to_title(seg):
    """Turn the current SRTSeg into Elastic Search dataformat.
    The return can be directly used to input into Search"""
    return {
        "original_index": seg.subtitle.index,
        "start": seg.start.total_seconds(),
        "end": seg.end.total_seconds(),
        "content": seg.subtitle.content,
        "sub_start": seg.subtitle.start.total_seconds(),
        "sub_end": seg.subtitle.end.total_seconds(),
    }


def titles_to_segs(hits):
    """Restore SRTSeg from Elastic Search Hits"""
    start = 0
    segs = []

    for hit in hits:
        seg = Seg()
        sub_duration = hit.sub_end - hit.sub_start
        seg_duration = hit.end - hit.start
        sub_offset = hit.sub_start - hit.start
        sub = srt.Subtitle(
            index=hit.original_index,
            start=timedelta(seconds=start + sub_offset),
            end=timedelta(seconds=start + sub_offset + sub_duration),
            content=hit.content,
        )
        seg.subtitle = sub
        seg.originalIndex = hit.original_index
        seg.selected = True
        seg.path = tools.COS_SERVER + "/" + hit.local_path
        seg.path = tools.low_res_url(seg.path)
        seg.start = timedelta(seconds=start)
        seg.end = timedelta(seconds=start + seg_duration)
        seg.duration = timedelta(seconds=seg_duration)
        segs.append(seg)
        start += seg_duration
    return segs
